from libcpp.vector cimport vector
import numpy as np

CPYTHON_STUFF = "ABNORMAL"

cdef extern from "mutual_info.h":
    double MutualInfo(vector[double] &, vector[double] &)
    vector[double] TDMI(vector[double] &, vector[double] &, vector[int] &)
    vector[double] TDMI_omp(vector[double] &, vector[double] &, vector[int] &)

def mutual_info(vec1, vec2):
    '''Mutual information estimator
    
    Args:
        vec1 (np.ndarray): first time series
        vec2 (np.ndarray): second time series

    Returns:
        float: mutual information value
    '''
    return MutualInfo(<vector[double]&> vec1, <vector[double]&> vec2)

def tdmi(vec1, vec2, n):
    '''Time-delayed Mutual information estimator
    
    Args:
        vec1 (np.ndarray): first time series
        vec2 (np.ndarray): second time series
        n (np.ndarray): number of delay steps. 
            Delay starts with 0.

    Returns:
        np.ndarray: time-delayed mutual information array
    '''
    return np.array(TDMI(<vector[double]&> vec1, <vector[double]&> vec2, <vector[int]&> n))

def tdmi_omp(vec1, vec2, n):
    '''Time-delayed Mutual information estimator (OpenMP accelerated version)
    
    Args:
        vec1 (np.ndarray): first time series
        vec2 (np.ndarray): second time series
        n (int): number of delay steps. Delay starts with 0.

    Returns:
        np.ndarray: time-delayed mutual information array
    '''
    return np.array(TDMI_omp(<vector[double]&> vec1, <vector[double]&> vec2, <vector[int]&> n))


import numpy as np

def mutual_info(x:np.ndarray, y:np.ndarray=None, rowvar=True)->np.ndarray:
    """Mutual Information Estimator for binary random variables.

    Args:
        x (np.ndarray): 2D numpy array of random variable. By default(rowvar),
            each row is a variable, each column is an observation.
        y (np.ndarray, optional): 2D array of random variable with the same
            format with x. If None, y is defined as a copy of x. 
            Defaults to None.
        rowvar (bool, optional): If True, each row of x(or y) is a variable,
            and each column is an observation. If False, each row of x(or y)
            is an observation, and each column is a variable. 
            Defaults to True.

    Raises:
        ValueError: x and y must have the same dimension of observations.

    Returns:
        np.ndarray: (nx, ny) mutual information matrix.
    """
    ref_flag = False
    if y is None:
        y = x   # shallow copy
        ref_flag = True
    else:
        if rowvar:
            if x.shape[1] != y.shape[1]:
                raise ValueError(f"x and y must have same second dimension, but "
                                 f"have shapes {x.shape} and {y.shape}")
        else:
            if x.shape[0] != y.shape[0]:
                raise ValueError(f"x and y must have same first dimension, but "
                                 f"have shapes {x.shape} and {y.shape}")
    if rowvar:
        N = x.shape[1]
        n00 = (~x).astype(np.float32) @ (~y.T).astype(np.float32)
        n01 = (~x).astype(np.float32) @ (y.T).astype(np.float32)
        if ref_flag:
            n10 =  n01.T
        else:
            n10 =  (x).astype(np.float32) @ (~y.T).astype(np.float32)
        # n11 =    x.astype(np.float32) @ (y.T).astype(np.float32)
        n11 =  np.ones_like(n00)*N - n00 - n01 - n10
        nx, ny = x.sum(1), y.sum(1)
    else:
        N = x.shape[0]
        n00 = (~x.T).astype(np.float32) @ (~y).astype(np.float32)
        n00 = (~x.T).astype(np.float32) @ (~y).astype(np.float32)
        n01 = (~x.T).astype(np.float32) @ y.astype(np.float32)
        if ref_flag:
            n10 =  n01.T
        else:
            n10 =  (x.T).astype(np.float32) @ (~y).astype(np.float32)
        # n11 =  (x.T).astype(np.float32) @ y.astype(np.float32)
        n11 =  np.ones_like(n00)*N - n00 - n01 - n10
        nx, ny = x.sum(0), y.sum(0)

    I = np.zeros_like(n00, dtype=float)
    for nxy_, nx_, ny_ in zip(
        (n00, n01, n10, n11),
        (N-nx, N-nx, nx, nx),
        (N-ny, ny, N-ny, ny)
    ):
        nonzero_mask = nxy_>0
        buffer = nxy_[nonzero_mask]*np.log(nxy_[nonzero_mask]/np.outer(nx_, ny_)[nonzero_mask])
        I[nonzero_mask] += buffer

    return I / N + np.log(N)



import setuptools
import sys
from Cython.Build import cythonize
import numpy
import sysconfig

extra_compile_args = sysconfig.get_config_var('CFLAGS')
if extra_compile_args is not None:
    extra_compile_args = extra_compile_args.split()
    extra_compile_args += ['-O2', '-std=c++11', '-fopenmp']
else:
    extra_compile_args = ['-O2', '-std=c++11', '-fopenmp']

ext_modules = [
    setuptools.Extension(
        "minfo.mi_c", 
        sources=["src/minfo.pyx", "src/mutual_info.cpp"], 
        language='c++',
        extra_compile_args= extra_compile_args,
            # + ['-fopenmp=libomp5' if sys.platform.startswith('darwin') else '-fopenmp'],
        include_dirs=['include/', numpy.get_include()]
    )
]

setuptools.setup(
    name="minfo",
    author="Kai Chen",
    author_email="kchen513@sjtu.edu.cn",

    version="0.0.6",
    url="https://github.com/NeoNeuron/minfo",

    description="Python-based mutual information estimator, designed for multiple data types.",

    install_requires=['numpy', 'cython', 'numba'],
    packages=setuptools.find_packages(),
    ext_modules = cythonize(ext_modules),
)


import numpy as np
from numba import njit, prange

from . import mi_c

def mutual_info(x:np.ndarray, y:np.ndarray, bins:int=None, algorithm:str='uniform')->np.ndarray:
    """_summary_

    Args:
        x (np.ndarray): 2D numpy array of random variable. By default(rowvar),
            each row is a variable, each column is an observation.
        y (np.ndarray): 2D array of random variable with the same
            format with x. 
        bins (int, optional): number of bins of histogram2d with uniform binning strategy.
            Defaults to None.
        algorithm (str, optional): algorithm to estimate PDFs, either 'uniform' or 'adaptive'.
            Defaults to 'uniform'.

    Raises:
        ValueError: x and y must have the same size of observations, same shape.

    Returns:
        np.ndarray: value of mutual information.
    """
    if x.shape[0] != y.shape[0]:
        raise ValueError(f"x and y must have same first dimension, but "
                             f"have shapes {x.shape} and {y.shape}")
    if bins is None:
        bins = int(np.sqrt(x.shape[0]))
    if algorithm == 'uniform':
        bins2 = np.array([bins, bins])
        mi = mutual_info_uni_nb(x, y, bins2)
    elif algorithm == 'adaptive':
        mi = mi_c.mutual_info(x, y)
    else:
        raise ValueError(f"{algorithm:s} is not acceptable. Please use either 'uniform' or 'adaptive'.")
    return mi

def TDMI(x:np.ndarray, y:np.ndarray, delay:int, bins:int=None, algorithm:str='uniform')->np.ndarray:
    """_summary_

    Args:
        x (np.ndarray): 2D numpy array of random variable. By default(rowvar),
            each row is a variable, each column is an observation.
        y (np.ndarray): 2D array of random variable with the same
            format with x. 
        delay (int | np.ndarray): number of delays, or explicit sequence of delay indices.
            Defaults to None.
        bins (int, optional): number of bins of histogram2d with uniform binning strategy.
            Defaults to None.
        algorithm (str, optional): algorithm to estimate PDFs, either 'uniform' or 'adaptive'.
            Defaults to 'uniform'.

    Raises:
        ValueError: x and y must have the same size of observations, same shape.

    Returns:
        np.ndarray: value of mutual information.
    """
    if x.shape[0] != y.shape[0]:
        raise ValueError(f"x and y must have same first dimension, but "
                             f"have shapes {x.shape} and {y.shape}")
    if bins is None:
        bins = int(np.sqrt(x.shape[0]))
    if isinstance(delay, int):
        delays = np.arange(delay)
    elif isinstance(delay, np.ndarray):
        delays = delay
    if algorithm == 'uniform':
        tdmi = TDMI_uni(x, y, delays, bins)
    elif algorithm == 'adaptive':
        tdmi = mi_c.tdmi_omp(x, y, delays)
    else:
        raise ValueError(f"{algorithm:s} is not acceptable. Please use either 'uniform' or 'adaptive'.")
    return tdmi

@njit(nogil=True, parallel=False)
def hist2d_numba_seq(data:np.ndarray, bins:np.ndarray, ranges:np.ndarray)->np.ndarray:
    """2D histogram estimation accelerated by Numba
        https://iscinumpy.gitlab.io/post/histogram-speeds-in-python/

    Args:
        data (np.ndarray): 2D array of data. Each row is a variable and each column is an observation.
        bins (np.ndarray): array of bins with shape (2,).
        ranges (np.ndarray): array of ranges with shape (2,2). ranges[i,:] is the range of i-th variable.

    Returns:
        np.ndarray: 2D array of histogram data.
    """
    H = np.zeros((bins[0], bins[1]), dtype=np.uint64)
    delta = 1 / ((ranges[:, 1] - ranges[:, 0]) / bins)

    for t in range(data.shape[1]):
        i = (data[0, t] - ranges[0, 0]) * delta[0]
        j = (data[1, t] - ranges[1, 0]) * delta[1]
        if 0 <= i < bins[0] and 0 <= j < bins[1]:
            H[int(i), int(j)] += 1

    return H

@njit(nogil=True, parallel=False)
def mutual_info_uni_nb(x:np.ndarray, y:np.ndarray, bins:np.ndarray)->float:
    """mutual information estimator. Joint PDF is estimated using uniform binning.

    Args:
        x (np.ndarray): 1D data
        y (np.ndarray): 1D data
        bins (np.ndarray | tuple | list): number of bins for histogram2d estimation.

    Returns:
        float: value of mutual information
    """
    N = x.shape[0]
    ranges = np.array([[x.min(), x.max()], [y.min(), y.max()]])
    dat = np.vstack((x,y))
    counts_xy = hist2d_numba_seq(dat, bins=bins, ranges=ranges)
    x_mar = counts_xy.sum(1)
    y_mar = counts_xy.sum(0)
    mi_val = 0.0
    for i in range(bins[0]):
        for j in range(bins[1]):
            if counts_xy[i,j] > 0:
                mi_val += counts_xy[i,j]*np.log(counts_xy[i,j]/(x_mar[i]*y_mar[j]))
    return mi_val / N + np.log(N)

# Python+Numpy based mutual information estimator with uniform binning.
def mutual_info_uni(x, y, bins):
    N = x.shape[0]
    counts_xy,_,_ = np.histogram2d(x, y, bins=bins,)
    x_mar = np.tile(counts_xy.sum(1, keepdims=True), (1, bins))
    y_mar = np.tile(counts_xy.sum(0, keepdims=True), (bins, 1))
    mi = np.nansum(counts_xy*np.log(counts_xy/(x_mar*y_mar)))
    return mi/N + np.log(N)

@njit(nogil=True, parallel=False)
def DMI_uni(x:np.ndarray, y:np.ndarray, delay:int, bins:int):
    bins2 = np.array([bins, bins])
    if delay == 0:
        return mutual_info_uni_nb(x, y, bins2)
    elif delay < 0:
        return mutual_info_uni_nb(x[-int(delay):],y[:int(delay)], bins2)
    elif delay > 0:
        return mutual_info_uni_nb(x[:-int(delay)],y[int(delay):], bins2)

@njit(nogil=True, parallel=True)
def TDMI_uni(x:np.ndarray, y:np.ndarray, time_range:np.ndarray, bins:np.ndarray=None)->np.ndarray:
    tdmi = np.zeros(len(time_range))
    for i in prange(len(time_range)):
        tdmi[i] = DMI_uni(x,y,time_range[i],bins)
    return tdmi
